﻿//License: GPLv3 (see LICENSE.txt)

//Last Release: 0.3.1
//WIP Version 0.3.2 a

using System.Collections.Generic;
using System.Collections;
using System.Linq;
using System;
using UnityEngine;

namespace Rotate
{
    [KSPAddon(KSPAddon.Startup.Flight, false)]
    public class VehicleRotator : MonoBehaviour
    {

        static bool started = false;
        static bool loaded = false;
        float lastrate = 1;
        static public Dictionary<string, Vector3> vessel_rotations = new Dictionary<string, Vector3>();  //Rotation zu jedem Vessel
        static public Dictionary<ProtoCrewMember, float> ArtificialGravity = new Dictionary<ProtoCrewMember, float>();
        private static String File
        {
            get { return KSPUtil.ApplicationRootPath + "/GameData/TimewarpRotate/VesselRotations_" + HighLogic.CurrentGame.Title + ".cfg"; }
        }
        int ticker = 0;

        protected Rect windowPos;
        bool minimize = true;
        string button = "maximize";
        bool unit = false;

        public void WindowGUI(int windowID)
        {
            GUIStyle mySty = new GUIStyle(GUI.skin.button);
            mySty.normal.textColor = mySty.focused.textColor = Color.white;
            mySty.hover.textColor = mySty.active.textColor = Color.yellow;
            mySty.onNormal.textColor = mySty.onFocused.textColor = mySty.onHover.textColor = mySty.onActive.textColor = Color.green;
            mySty.padding = new RectOffset(8, 8, 8, 8);

            GUILayout.BeginVertical();
            if (GUILayout.Button(button, mySty, GUILayout.ExpandWidth(true)))//GUILayout.Button is "true" when clicked
            {
                minimize = !minimize;
                if (minimize)
                {
                    button = "maximize";
                    windowPos.height = 10;
                    RenderingManager.AddToPostDrawQueue(3, new Callback(drawGUI));//start the GUI  
                }
                else
                {
                    button = "minimize";
                    RenderingManager.AddToPostDrawQueue(3, new Callback(drawGUI));//start the GUI  
                }
            }
            if (!minimize)
            {
                if (GUILayout.Button("change units", mySty, GUILayout.ExpandWidth(true)))//GUILayout.Button is "true" when clicked
                {
                    unit = !unit;
                }

                CalculateArtificialGeForce();

	            foreach(ProtoCrewMember kerb in FlightGlobals.ActiveVessel.GetVesselCrew())
                {
                    if (kerb.KerbalRef.state != Kerbal.States.DEAD)
                    {
                        GUILayout.BeginHorizontal();
                        if (ArtificialGravity[kerb] > 10 * 9.81)
                        {
                            GUILayout.Label("(!!!) ");
                        }
                        GUILayout.Label(kerb.name.TrimEnd("Kerman".ToCharArray()));
                        GUILayout.FlexibleSpace();
                        if (unit)
                        {
                            ArtificialGravity[kerb] = (float)((int)(ArtificialGravity[kerb] * 10)) / 10;
                            GUILayout.Label(ArtificialGravity[kerb].ToString() + " m/s");
                        }
                        else
                        {
                            ArtificialGravity[kerb] = (float)((int)((ArtificialGravity[kerb] / 9.81) * 10)) / 10;
                            GUILayout.Label(ArtificialGravity[kerb].ToString() + " G");
                        }

                        print(kerb.KerbalRef.stupidity.ToString());

                        GUILayout.EndHorizontal();
                    }
                }
        }

            GUILayout.EndVertical();

            //DragWindow makes the window draggable. The Rect specifies which part of the window it can by dragged by, and is 
            //clipped to the actual boundary of the window. You can also pass no argument at all and then the window can by
            //dragged by any part of it. Make sure the DragWindow command is AFTER all your other GUI input stuff, or else
            //it may "cover up" your controls and make them stop responding to the mouse.
            GUI.DragWindow(new Rect(0, 0, 10000, 20));

        }
        private void drawGUI()
        {
            GUI.skin = HighLogic.Skin;
            windowPos = GUILayout.Window(1, windowPos, WindowGUI, "Artificial Gravity", GUILayout.MinWidth(10));
        }

        public void CleanVesselRotations()
        {
            Dictionary<string, Vector3> temp_vessel_rotations = new Dictionary<string, Vector3>();

            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                temp_vessel_rotations[vessel.id.ToString()] = Vector3.zero;
            }
            foreach (KeyValuePair<string, Vector3> entry in vessel_rotations)
            {
                temp_vessel_rotations[entry.Key] = entry.Value;
            }

            vessel_rotations.Clear();

            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                vessel_rotations[vessel.id.ToString()] = temp_vessel_rotations[vessel.id.ToString()];
            }

            temp_vessel_rotations.Clear();
        }
        public void Save()
        {
            try
            {
                CleanVesselRotations();

                ConfigNode save = new ConfigNode();
                //ConfigNode.CreateConfigFromObject(this, 0, save); (?)
                //ConfigNode ships = new ConfigNode("SHIPS"); (?)
                foreach (KeyValuePair<string, Vector3> entry in vessel_rotations)
                {
                    save.AddValue(entry.Key, KSPUtil.WriteVector(entry.Value));
                }
                //save.AddNode(ships); (?)

                save.Save(File);
            }
            catch (Exception e) { print("[Rotate]: An error occurred while attempting to save: " + e.Message); }
        }
        public static void Load()
        {
            loaded = true;
            ConfigNode load = ConfigNode.Load(File);
            if (load == null)
            {
                print("[Rotate]: null save");
                return;
            }
            vessel_rotations.Clear();
            //Prefill it with empty Vectors
            foreach (Vessel vprime in FlightGlobals.Vessels)
            {
                vessel_rotations[vprime.id.ToString()] = Vector3.zero;
            }
            //Fill it with the saved ones
            foreach (ConfigNode.Value s in load.values)
            {
                vessel_rotations[s.name] = KSPUtil.ParseVector3(s.value);
            }
        }

        void OnVesselGoOnRails(Vessel vprime)
        {
            if (vprime.ActionGroups[KSPActionGroup.SAS])
            {
                vessel_rotations[vprime.id.ToString()] = Vector3.zero;
            }
            else
            {
                vessel_rotations[vprime.id.ToString()] = vprime.angularVelocity;
            }

            Save();

        }
        void OnVesselGoOffRails(Vessel vprime)
        {
            print("[Rotate]: Vessel " + vprime.vesselName.ToString() + " now going off Rails");

            Load();

            try
            {
                Vector3 av = vessel_rotations[vprime.id.ToString()];

                Vector3 COM = vprime.findWorldCenterOfMass();
                foreach (Part p in vprime.parts) //Applying force on every part
                {
                    try
                    {
                        if (p.rigidbody == null) continue;
                        p.rigidbody.AddTorque(vprime.transform.rotation * av, ForceMode.VelocityChange);
                        p.rigidbody.AddForce(Vector3.Cross(vprime.transform.rotation * av, (p.rigidbody.position - COM)), ForceMode.VelocityChange);
                    }
                    catch (NullReferenceException nre)
                    {
                        print("Got NRE: " + nre.Message);
                    }

                }
            }
            catch (Exception e)
            {
                print("Caugt exception: " + e.Message);
                // Don't really need to do anything; we just haven't met this vehicle yet.
            }

        }

        void Start()
        {
            if (!started)
            {
                GameEvents.onVesselGoOnRails.Add(OnVesselGoOnRails);
                GameEvents.onVesselGoOffRails.Add(OnVesselGoOffRails);
                GameEvents.onGamePause.Add(OnPause);
                started = true;
            }

            //Prefill vessel_rotations with empty Vectors, for first time loading
            foreach (Vessel vprime in FlightGlobals.Vessels)
            {
               vessel_rotations[vprime.id.ToString()] = Vector3.zero;
            }

            RenderingManager.AddToPostDrawQueue(3, new Callback(drawGUI));//start the GUI
            if ((windowPos.x == 0) && (windowPos.y == 0))//windowPos is used to position the GUI window, lets set it in the center of the screen
            {
                windowPos = new Rect(Screen.width / 4, 0, 150, 10);
            }
        }
        void OnPause()
        {
            foreach (Vessel vessel in FlightGlobals.Vessels)
            {
                if(!vessel.packed)
                {
                    vessel_rotations[vessel.id.ToString()] = vessel.angularVelocity;
                }
            }
            Save();
        }

        public void RotateWhilePacked(Vessel vessel)
        {
            Vector3 angvel = vessel_rotations[vessel.id.ToString()];

            if (vessel.packed == true && vessel.situation != Vessel.Situations.LANDED
                                      && vessel.situation != Vessel.Situations.PRELAUNCH
                                      && vessel.situation != Vessel.Situations.DOCKED
                                      && vessel.situation != Vessel.Situations.SPLASHED)
            {
                    float angle = angvel.magnitude * TimeWarp.CurrentRate;
                    Quaternion rotation = Quaternion.AngleAxis(angle, angvel);
                    vessel.SetRotation(vessel.GetTransform().rotation * rotation);

                //TODO: SAS Modes
            }
        }
        public void CalculateArtificialGeForce()
        {  
            foreach (ProtoCrewMember kerb in FlightGlobals.ActiveVessel.GetVesselCrew())
            {
                float speed = FlightGlobals.ActiveVessel.angularVelocity.magnitude * (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;
                float force = (speed * speed) / (kerb.seat.part.rigidbody.position - FlightGlobals.ActiveVessel.findWorldCenterOfMass()).magnitude;

                ArtificialGravity[kerb] = force;

                if (force > 9.81 * 22 - 9.81 * kerb.KerbalRef.stupidity * 4)
                {
                    kerb.KerbalRef.die(); //Let Kerbal die if GeForce exceeds ~20 G's
                    ArtificialGravity.Remove(kerb);
                    windowPos.height = 10;
                }
            }
        }
        void FixedUpdate()
        {
            ticker++;
            if (ticker > 50)
            {
                ticker = 0;
                RenderingManager.AddToPostDrawQueue(3, new Callback(drawGUI));//start the GUI
            }
            if (!loaded)
            {
                Load();
            }
            try
            {
                foreach (Vessel vessel in FlightGlobals.Vessels)
                {
                    RotateWhilePacked(vessel);
                }
                lastrate = TimeWarp.CurrentRate;
            }
            catch (NullReferenceException nre)
            {
                print("[Rotate]: NullReference: " + nre.ToString());
            }
        }
    }
}

